"""Constants for the Bosch SHC integration."""

CONF_HOSTNAME = "hostname"
CONF_SHC_CERT = "bosch_shc-cert.pem"
CONF_SHC_KEY = "bosch_shc-key.pem"
CONF_SSL_CERTIFICATE = "ssl_certificate"
CONF_SSL_KEY = "ssl_key"

DATA_SESSION = "session"
DATA_POLLING_HANDLER = "polling_handler"

DOMAIN = "bosch_shc"
