"""Constants for Bravia TV integration."""
from __future__ import annotations

from typing import Final

ATTR_CID: Final = "cid"
ATTR_MAC: Final = "macAddr"
ATTR_MANUFACTURER: Final = "Sony"
ATTR_MODEL: Final = "model"

CONF_IGNORED_SOURCES: Final = "ignored_sources"

BRAVIA_CONFIG_FILE: Final = "bravia.conf"
CLIENTID_PREFIX: Final = "HomeAssistant"
DEFAULT_NAME: Final = f"{ATTR_MANUFACTURER} Bravia TV"
DOMAIN: Final = "braviatv"
NICKNAME: Final = "Home Assistant"
