"""Constants for Brother integration."""
from __future__ import annotations

from typing import Final

DATA_CONFIG_ENTRY: Final = "config_entry"

DOMAIN: Final = "brother"

PRINTER_TYPES: Final = ["laser", "ink"]

SNMP: Final = "snmp"
