"""HTTP specific constants."""
from typing import Final

KEY_AUTHENTICATED: Final = "ha_authenticated"
KEY_HASS: Final = "hass"
KEY_HASS_USER: Final = "hass_user"
KEY_HASS_REFRESH_TOKEN_ID: Final = "hass_refresh_token_id"
