"""Shared constants for script and automation tracing and debugging."""

CONF_STORED_TRACES = "stored_traces"
DATA_TRACE = "trace"
DATA_TRACE_STORE = "trace_store"
DATA_TRACES_RESTORED = "trace_traces_restored"
DEFAULT_STORED_TRACES = 5  # Stored traces per script or automation
